"""
  TheengsGateway - Decode things and devices and publish data to an MQTT broker

    Copyright: (c)Florian ROBERT

    This file is part of TheengsGateway.

    TheengsGateway is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TheengsGateway is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# python 3.6
# encoding=utf8

import json
import re
from .ble_gateway import gateway, logger
from ._decoder import getProperties

ha_dev_classes = ["battery",
                  "carbon_monoxide",
                  "carbon_dioxide",
                  "humidity",
                  "illuminance",
                  "signal_strength",
                  "temperature",
                  "timestamp",
                  "pressure",
                  "power",
                  "current",
                  "energy",
                  "power_factor",
                  "voltage"]

ha_dev_units = ["W",
                "kW",
                "V",
                "A",
                "W",
                "°C",
                "°F",
                "ms",
                "s",
                "hPa",
                "kg",
                "lb",
                "µS/cm",
                "lx",
                "%",
                "dB",
                "B"]


class discovery(gateway):
    def __init__(self, broker, port, username, password, adapter, scanning_mode,
                 discovery_topic, discovery_device_name, discovery_filter):
        super().__init__(broker, port, username, password, adapter, scanning_mode)
        self.discovery_topic = discovery_topic
        self.discovery_device_name = discovery_device_name
        self.discovered_entities = []
        self.discovery_filter = discovery_filter

    def connect_mqtt(self):
        super().connect_mqtt()

    def publish(self, msg, pub_topic, retain=False):
        return super().publish(msg, pub_topic, retain)

    # publish sensor directly to home assistant via mqtt discovery
    def publish_device_info(self, pub_device):
        pub_device_uuid = pub_device['id'].replace(':', '')
        device_data = json.dumps(pub_device)
        if (pub_device_uuid in self.discovered_entities or
                pub_device['model_id'] in self.discovery_filter):
            logger.debug("Already discovered or filtered: %s" %
                         pub_device_uuid)
            self.publish(device_data, self.pub_topic + '/' +
                         pub_device_uuid)
            return

        logger.info(f"publishing device `{pub_device}`")
        pub_device['properties'] = json.loads(
            getProperties(pub_device['model_id']))['properties']

        hadevice = {}
        hadevice['identifiers'] = list({pub_device_uuid})
        hadevice['connections'] = [list(('mac', pub_device_uuid))]
        hadevice['manufacturer'] = pub_device['brand']
        hadevice['model'] = pub_device['model_id']
        if 'name' in pub_device:
            hadevice['name'] = pub_device['name']
        else:
            hadevice['name'] = pub_device['model']
        hadevice['via_device'] = self.discovery_device_name

        discovery_topic = self.discovery_topic + "/" + pub_device_uuid
        state_topic = self.pub_topic + "/" + pub_device_uuid
        state_topic = re.sub(r'.+?/', '+/', state_topic,
                             len(re.findall(r'/', state_topic)) - 1)
        data = getProperties(pub_device['model_id'])
        data = json.loads(data)
        data = data['properties']

        for k in data.keys():
            device = {}
            device['stat_t'] = state_topic
            if k in pub_device['properties']:
                if pub_device['properties'][k]['name'] in ha_dev_classes:
                    device['dev_cla'] = pub_device['properties'][k]['name']
                if pub_device['properties'][k]['unit'] in ha_dev_units:
                    device['unit_of_meas'] = pub_device['properties'][k]['unit']
            device['name'] = pub_device['model_id'] + "-" + k
            device['uniq_id'] = pub_device_uuid + "-" + k
            device['val_tpl'] = "{{ value_json." + k + " | is_defined }}"
            device['state_class'] = "measurement"
            config_topic = discovery_topic + "-" + k + "/config"
            device['device'] = hadevice
            self.publish(json.dumps(device), config_topic, True)

        self.discovered_entities.append(pub_device_uuid)
        self.publish(device_data, self.pub_topic + '/' +
                     pub_device_uuid)




"""
  TheengsGateway - Decode things and devices and publish data to an MQTT broker

    Copyright: (c)Florian ROBERT

    This file is part of TheengsGateway.

    TheengsGateway is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TheengsGateway is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
import os
import json
import argparse
from .ble_gateway import run

default_config = {
    "host":"",
    "port":1883,
    "user":"",
    "pass":"",
    "ble_scan_time":5,
    "ble_time_between_scans":5,
    "publish_topic": "home/TheengsGateway/BTtoMQTT",
    "subscribe_topic": "home/TheengsGateway/+",
    "log_level": "WARNING",
    "discovery": 1,
    "discovery_topic": "homeassistant/sensor",
    "discovery_device_name": "TheengsGateway",
    "discovery_filter": ["IBEACON", "GAEN", "MS-CDP"],
    "adapter": "",
    "scanning_mode": "active"
}

conf_path = os.path.expanduser('~') + '/theengsgw.conf'

parser =  argparse.ArgumentParser()
parser.add_argument('-H', '--host', dest='host', type=str, help="MQTT host address")
parser.add_argument('-P', '--port', dest='port', type=int, help="MQTT host port")
parser.add_argument('-u', '--user', dest='user', type=str, help="MQTT username")
parser.add_argument('-p', '--pass', dest='pwd', type=str, help="MQTT password")
parser.add_argument('-pt', '--pub_topic', dest='pub_topic', type=str, help="MQTT publish topic")
parser.add_argument('-st', '--sub_topic', dest='sub_topic', type=str, help="MQTT subscribe topic")
parser.add_argument('-pa', '--publish_all', dest='publish_all', type=int, help="Enable(1) or disable(0) publishing of all beacons")
parser.add_argument('-sd', '--scan_duration', dest='scan_dur', type=int, help="BLE scan duration (seconds)")
parser.add_argument('-tb', '--time_between', dest='time_between', type=int, help="Seconds to wait between scans")
parser.add_argument('-ll', '--log_level', dest='log_level', type=str, help="TheengsGateway log level",
                    choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])
parser.add_argument('-Dt', '--discovery-topic', dest='discovery_topic', type=str, help="MQTT Discovery topic")
parser.add_argument('-D', '--discovery', dest='discovery', type=int, help="Enable(1) or disable(0) MQTT discovery")
parser.add_argument('-Dn', '--discovery_name', dest='discovery_device_name', type=str, help="Device name for Home Assistant")
parser.add_argument('-Df', '--discovery_filter', dest='discovery_filter', nargs='+', default=[],
                    help="Device discovery filter list for Home Assistant")
parser.add_argument('-a', '--adapter', dest='adapter', type=str, help="Bluetooth adapter (e.g. hci1 on Linux)")
parser.add_argument('-s', '--scanning_mode', dest='scanning_mode', type=str, choices=("active", "passive"), help="Scanning mode (default: active)")
args = parser.parse_args()

try:
    with open(conf_path, 'r') as config_file:
        config = json.load(config_file)
except:
    config = default_config

# Merge default configuration, with data read from the configuration file overriding default data.
# This guarantees that all keys we refer to are in the dictionary.
config = {**default_config, **config}

if args.host:
    config['host'] = args.host
if args.port:
    config['port'] = args.port
if args.user:
    config['user'] = args.user
if args.pwd:
    config['pass'] = args.pwd
if args.pub_topic:
    config['publish_topic'] = args.pub_topic
if args.sub_topic:
    config['subscribe_topic'] = args.sub_topic
if args.publish_all:
    config['publish_all'] = args.publish_all
if args.scan_dur:
    config['ble_scan_time'] = args.scan_dur
if args.time_between:
    config['ble_time_between_scans'] = args.time_between
if args.log_level:
    config['log_level'] = args.log_level

if args.discovery is not None:
    config['discovery'] = args.discovery
elif not 'discovery' in config.keys():
    config['discovery'] = default_config['discovery']
    config['discovery_topic'] = default_config['discovery_topic']
    config['discovery_device_name'] = default_config['discovery_device_name']
    config['discovery_filter'] = default_config['discovery_filter']

if args.discovery_topic:
    config['discovery_topic'] = args.discovery_topic
elif not 'discovery_topic' in config.keys():
    config['discovery_topic'] = default_config['discovery_topic']

if args.discovery_device_name:
    config['discovery_device_name'] = args.discovery_device_name
elif not 'discovery_device_name' in config.keys():
    config['discovery_device_name'] = default_config['discovery_device_name']

if args.discovery_filter:
    config['discovery_filter'] = default_config['discovery_filter']
    if args.discovery_filter[0] != "reset":
        for item in args.discovery_filter:
            config['discovery_filter'].append(item)
elif not 'discovery_filter' in config.keys():
    config['discovery_filter'] = default_config['discovery_filter']

if args.adapter:
    config['adapter'] = args.adapter

if args.scanning_mode:
    config['scanning_mode'] = args.scanning_mode

if not config['host']:
    sys.exit('Invalid MQTT host')

try:
    with open(conf_path, 'w') as config_file:
        config_file.write(json.dumps(config, sort_keys=True, indent=4))
except:
    raise SystemExit('Unable to open config file')

run(conf_path)



"""
  TheengsGateway - Decode things and devices and publish data to an MQTT broker

    Copyright: (c)Florian ROBERT

    This file is part of TheengsGateway.

    TheengsGateway is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TheengsGateway is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# python 3.6

import asyncio
from datetime import datetime
import json
import struct
import sys
import logging
import platform
from random import randrange
import time

from bleak import BleakClient, BleakError, BleakScanner
from ._decoder import decodeBLE, getProperties, getAttribute
from paho.mqtt import client as mqtt_client
from threading import Thread

if platform.system() == "Linux":
    from bleak.assigned_numbers import AdvertisementDataType
    from bleak.backends.bluezdbus.advertisement_monitor import OrPattern
    from bleak.backends.bluezdbus.scanner import BlueZScannerArgs

SECONDS_IN_HOUR = 3600
SECONDS_IN_DAY = 86400

LYWSD02_TIME_UUID = "ebe0ccb7-7a0a-4b0c-8a1a-6ff2997da3a6"

logger = logging.getLogger('BLEGateway')

class gateway:
    def __init__(self, broker, port, username, password, adapter, scanning_mode):
        self.broker = broker
        self.port = port
        self.username = username
        self.password = password
        self.adapter = adapter
        self.scanning_mode = scanning_mode
        self.stopped = False
        self.lywsd02_updates = {}

    def connect_mqtt(self):
        def on_connect(client, userdata, flags, rc):
            if rc == 0:
                logger.info("Connected to MQTT Broker!")
                self.subscribe(self.sub_topic)
            else:
                logger.error(f"Failed to connect to MQTT broker %s:%d rc: %d" % (self.broker, self.port, rc))
                self.client.connect(self.broker, self.port)

        def on_disconnect(client, userdata,rc=0):
            logger.error(f"Disconnected rc = %d" % (rc))

        self.client = mqtt_client.Client()
        self.client.username_pw_set(self.username, self.password)
        self.client.on_connect = on_connect
        self.client.on_disconnect = on_disconnect
        try:
            self.client.connect(self.broker, self.port)
        except:
            pass

    def subscribe(self, sub_topic):
        def on_message(client_, userdata, msg):
            logger.info(f"Received `{msg.payload.decode()}` from `{msg.topic}` topic")
            try:
                msg_json = json.loads(str(msg.payload.decode()))
            except:
                return
            address = msg_json["id"]
            decoded_json = decodeBLE(json.dumps(msg_json))
            if decoded_json:
                if gw.discovery:
                    gw.publish_device_info(json.loads(decoded_json)) ## publish sensor data to home assistant mqtt discovery
                else:
                    gw.publish(decoded_json, gw.pub_topic + '/' + address.replace(':', ''))
            elif gw.publish_all:
                gw.publish(str(msg.payload.decode()), gw.pub_topic + '/' + address.replace(':', ''))

        self.client.subscribe(sub_topic)
        self.client.on_message = on_message
        logger.info(f"Subscribed to {sub_topic}")


    def publish(self, msg, pub_topic=None, retain=False):
        if not pub_topic:
            pub_topic = self.pub_topic

        result = self.client.publish(pub_topic, msg, 0, retain)
        status = result[0]
        if status == 0:
            logger.info(f"Sent `{msg}` to topic `{pub_topic}`")
        else:
            logger.error(f"Failed to send message to topic {pub_topic}")

    def add_lywsd02(self, address, decoded_json):
        if json.loads(decoded_json)["model_id"] == "LYWSD02":
            if address not in self.lywsd02_updates:
                # Add a random time in the last day as a starting point for the daily update.
                # This prevents the gateway from connecting to all devices at the same time.
                self.lywsd02_updates[address] = datetime.now().timestamp() - randrange(SECONDS_IN_DAY)
                logger.info(f"Found LYWSD02 device {address}, synchronizing time daily...")

    async def update_lywsd02_time(self):
        for address, timestamp in self.lywsd02_updates.copy().items():
            if datetime.now().timestamp() - timestamp > SECONDS_IN_DAY:
                logger.info(f"Synchronizing time for LYWSD02 device {address}...")
                try:
                    async with BleakClient(address) as lywsd02_client:
                        # Set timezone offset
                        if time.daylight:
                            timezone_offset = -time.altzone // SECONDS_IN_HOUR
                        else:
                            timezone_offset = -time.timezone // SECONDS_IN_HOUR

                        # Pack data for current time and timezone
                        lywsd02_time = struct.pack('Ib', int(datetime.now().timestamp()), timezone_offset)

                        # Write time and timezone to device
                        await lywsd02_client.write_gatt_char(LYWSD02_TIME_UUID, lywsd02_time)
                        logger.info(f"Synchronized time for LYWSD02 device {address}.")
                        # Reset timestamp to synchronize again in a day
                        self.lywsd02_updates[address] = datetime.now().timestamp()
                except BleakError as e:
                    logger.error(e)
                    del self.lywsd02_updates[address]
                except asyncio.exceptions.TimeoutError:
                    logger.error(f"Can't connect to LYWSD02 device {address}.")
                    del self.lywsd02_updates[address]

    async def ble_scan_loop(self):
        scanner_kwargs = {"scanning_mode": self.scanning_mode}

        # Passive scanning with BlueZ needs at least one or_pattern.
        # The following matches all devices.
        if platform.system() == "Linux" and self.scanning_mode == "passive":
            scanner_kwargs["bluez"] = BlueZScannerArgs(
                or_patterns=[
                    OrPattern(0, AdvertisementDataType.FLAGS, b"\x06"),
                    OrPattern(0, AdvertisementDataType.FLAGS, b"\x1a"),
                ]
            )

        if self.adapter:
            scanner_kwargs["adapter"] = self.adapter

        scanner = BleakScanner(**scanner_kwargs)
        scanner.register_detection_callback(self.detection_callback)
        logger.info('Starting BLE scan')
        self.running = True
        while not self.stopped:
            try:
                if self.client.is_connected():
                    await scanner.start()
                    await asyncio.sleep(self.scan_time)
                    await scanner.stop()
                    await asyncio.sleep(self.time_between_scans)

                    # Update time for all LYWSD02 devices once a day
                    await self.update_lywsd02_time()
                else:
                    await asyncio.sleep(5.0)
            except Exception as e:
                raise e

        logger.error('BLE scan loop stopped')
        self.running = False

    def detection_callback(self, device, advertisement_data):
        logger.debug("%s RSSI:%d %s" % (device.address, device.rssi, advertisement_data))
        data_json = {}

        if advertisement_data.service_data:
            dstr = list(advertisement_data.service_data.keys())[0]
            data_json['servicedatauuid'] = dstr[4:8]
            dstr = str(list(advertisement_data.service_data.values())[0].hex())
            data_json['servicedata'] = dstr

        if advertisement_data.manufacturer_data:
            dstr = str(struct.pack('<H', list(advertisement_data.manufacturer_data.keys())[0]).hex())
            dstr += str(list(advertisement_data.manufacturer_data.values())[0].hex())
            data_json['manufacturerdata'] = dstr

        if advertisement_data.local_name:
            data_json['name'] = advertisement_data.local_name

        if data_json:
            data_json['id'] = device.address
            data_json['rssi'] = device.rssi
            decoded_json = decodeBLE(json.dumps(data_json))

            if decoded_json:
                if gw.discovery:
                    gw.publish_device_info(json.loads(decoded_json)) ## publish sensor data to home assistant mqtt discovery
                else:
                    gw.publish(decoded_json, gw.pub_topic + '/' + device.address.replace(':', ''))

                # Add new LYWSD02 devices to dictionary of devices to synchronize time
                self.add_lywsd02(device.address, decoded_json)
            elif gw.publish_all:
                gw.publish(json.dumps(data_json), gw.pub_topic + '/' + device.address.replace(':', ''))

def run(arg):
    global gw

    try:
        with open(arg) as config_file:
            config = json.load(config_file)
    except:
        raise SystemExit(f"Invalid File: {sys.argv[1]}")

    log_level = config.get("log_level", "WARNING").upper()
    if log_level == "DEBUG":
        log_level = logging.DEBUG
    elif log_level == "INFO":
        log_level = logging.INFO
    elif log_level == "WARNING":
        log_level = logging.WARNING
    elif log_level == "ERROR":
        log_level = logging.ERROR
    elif log_level == "CRITICAL":
        log_level = logging.CRITICAL
    else:
        log_level = logging.WARNING

    if config['discovery']:
        from .discovery import discovery
        gw = discovery(config["host"], int(config["port"]), config["user"],
                       config["pass"], config["adapter"], config["scanning_mode"],
                       config["discovery_topic"], config["discovery_device_name"],
                       config["discovery_filter"])
    else:
        try:
          gw = gateway(config["host"], int(config["port"]), config["user"], config["pass"], config["adapter"], config["scanning_mode"])
        except:
          raise SystemExit(f"Missing or invalid MQTT host parameters")

    gw.discovery = config['discovery']
    gw.scan_time = config.get("ble_scan_time", 5)
    gw.time_between_scans = config.get("ble_time_between_scans", 0)
    gw.sub_topic = config.get("subscribe_topic", "gateway_sub")
    gw.pub_topic = config.get("publish_topic", "gateway_pub")
    gw.publish_all = config.get("publish_all", 5)

    logging.basicConfig()
    logger.setLevel(log_level)

    loop = asyncio.get_event_loop()
    t = Thread(target=loop.run_forever, daemon=True)
    t.start()
    asyncio.run_coroutine_threadsafe(gw.ble_scan_loop(), loop)

    gw.connect_mqtt()

    try:
        gw.client.loop_forever()
    except(KeyboardInterrupt, SystemExit):
        gw.client.disconnect()
        gw.stopped = True
        while gw.running:
            pass
        loop.call_soon_threadsafe(loop.stop)
        t.join()

if __name__ == '__main__':
    try:
        arg = sys.argv[1]
    except IndexError:
        raise SystemExit(f"Usage: {sys.argv[0]} /path/to/config_file")
    run(arg)
